/// Compute how different two floats are in ulps.
///
/// Notes:
/// - Treats 0.0 and -0.0 as zero ulps apart, and extends the
///   implications of that to the rest of the numbers.  E.g. the numbers
///   just before and after 0.0/-0.0 are only two ulps apart, not three.
/// - Infinity is one ulp past float max, and converse for -infinity.
/// - If either number is NaN, returns `u32::MAX`.
#[inline(always)]
pub fn ulp_diff(a: f32, b: f32) -> u32 {
    const SIGN_BIT: u32 = 1 << 31;
    const INFINITY: u32 = 0x7f800000;

    let a = a.to_bits();
    let b = b.to_bits();

    let a_sign = a & SIGN_BIT;
    let b_sign = b & SIGN_BIT;
    let a_abs = a & !SIGN_BIT;
    let b_abs = b & !SIGN_BIT;

    if a_abs > INFINITY || b_abs > INFINITY {
        // NaNs always return maximum ulps apart.
        u32::MAX
    } else if a_sign == b_sign {
        a_abs.max(b_abs) - a_abs.min(b_abs)
    } else {
        a_abs + b_abs
    }
}

/// Checks if two floats are approximately equal, within `max_ulps`.
#[inline(always)]
pub fn ulps_eq(a: f32, b: f32, max_ulps: u32) -> bool {
    // The minimum ensures that NaNs never return true.
    ulp_diff(a, b) <= max_ulps.min(u32::MAX - 1)
}

/// Increments to the next representable floating point number.
///
/// Notes:
/// - 0.0 and -0.0 are treated as the same value.  E.g. starting from the
///   number just before -0.0, it only takes two increments to get to the
///   number just after 0.0.
/// - Infinity, NaN, and their negative counterparts are returned
///   unchanged.
/// - Incrementing `f32::MAX` results in infinity.
#[inline(always)]
pub fn increment_ulp(v: f32) -> f32 {
    if v.is_finite() {
        if v > 0.0 {
            f32::from_bits(v.to_bits() + 1)
        } else if v < -0.0 {
            f32::from_bits(v.to_bits() - 1)
        } else {
            f32::from_bits(1)
        }
    } else {
        // Infinity or NaN.
        v
    }
}

/// Decrements to the previous representable floating point number.
///
/// Notes:
/// - 0.0 and -0.0 are treated as the same value.  E.g. starting from the
///   number just after 0.0, it only takes two decrements to get to the
///   number just before -0.0.
/// - Infinity, NaN, and their negative counterparts are returned
///   unchanged.
/// - Decrementing `-f32::MAX` results in -infinity.
#[inline(always)]
pub fn decrement_ulp(v: f32) -> f32 {
    if v.is_finite() {
        if v > 0.0 {
            f32::from_bits(v.to_bits() - 1)
        } else if v < -0.0 {
            f32::from_bits(v.to_bits() + 1)
        } else {
            f32::from_bits(0x80000001)
        }
    } else {
        // Infinity or NaN.
        v
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn ulp_diff_test() {
        assert_eq!(ulp_diff(1.0, 1.0), 0);
        assert_eq!(ulp_diff(0.0, 0.0), 0);
        assert_eq!(ulp_diff(0.0, -0.0), 0);
        assert_eq!(ulp_diff(1.0, 2.0), 1 << 23);
        assert_eq!(ulp_diff(2.0, 4.0), 1 << 23);
        assert_eq!(ulp_diff(-1.0, -2.0), 1 << 23);
        assert_eq!(ulp_diff(-2.0, -4.0), 1 << 23);
        assert_eq!(ulp_diff(-1.0, 1.0), 0x7f000000);
        assert_eq!(ulp_diff(0.0, 1.0), 0x3f800000);
        assert_eq!(ulp_diff(-0.0, 1.0), 0x3f800000);
        assert_eq!(ulp_diff(0.0, -1.0), 0x3f800000);
        assert_eq!(ulp_diff(-0.0, -1.0), 0x3f800000);
        assert_eq!(ulp_diff(f32::INFINITY, -f32::INFINITY), 0xff000000);
        assert_eq!(ulp_diff(f32::NAN, f32::NAN), 0xffffffff);
        assert_eq!(ulp_diff(f32::NAN, 1.0), 0xffffffff);
        assert_eq!(ulp_diff(1.0, f32::NAN), 0xffffffff);
        assert_eq!(ulp_diff(-f32::NAN, 1.0), 0xffffffff);
        assert_eq!(ulp_diff(1.0, -f32::NAN), 0xffffffff);
        assert_eq!(ulp_diff(0.0, f32::from_bits(0.0f32.to_bits() + 1)), 1);
        assert_eq!(ulp_diff(-0.0, f32::from_bits(0.0f32.to_bits() + 1)), 1);
    }

    #[test]
    fn ulps_eq_test() {
        assert!(ulps_eq(1.0, 1.0, 0));
        assert!(ulps_eq(1.0, 1.0, 1));
        assert!(ulps_eq(0.0, 0.0, 0));
        assert!(ulps_eq(0.0, -0.0, 0));

        assert!(ulps_eq(1.0, 2.0, 1 << 23));
        assert!(!ulps_eq(1.0, 2.0, (1 << 23) - 1));

        assert!(ulps_eq(0.0, f32::from_bits(0.0f32.to_bits() + 1), 1));
        assert!(!ulps_eq(0.0, f32::from_bits(0.0f32.to_bits() + 1), 0));

        assert!(ulps_eq(-0.0, f32::from_bits(0.0f32.to_bits() + 1), 1));
        assert!(!ulps_eq(-0.0, f32::from_bits(0.0f32.to_bits() + 1), 0));

        assert!(ulps_eq(std::f32::INFINITY, -std::f32::INFINITY, 0xff000000));
        assert!(!ulps_eq(
            std::f32::INFINITY,
            -std::f32::INFINITY,
            0xff000000 - 1
        ));

        assert!(!ulps_eq(std::f32::NAN, std::f32::NAN, 0));
        assert!(!ulps_eq(-std::f32::NAN, -std::f32::NAN, 0));
        assert!(!ulps_eq(std::f32::NAN, std::f32::NAN, u32::MAX));
        assert!(!ulps_eq(std::f32::NAN, std::f32::INFINITY, 1 << 31));
        assert!(!ulps_eq(std::f32::INFINITY, std::f32::NAN, 1 << 31));
    }

    #[test]
    fn inc_ulp() {
        assert!(increment_ulp(1.0) > 1.0);
        assert!(increment_ulp(-1.0) > -1.0);
        assert!(increment_ulp(0.0) > 0.0);
        assert!(increment_ulp(0.0) > -0.0);
        assert!(increment_ulp(-0.0) > 0.0);
        assert!(increment_ulp(-0.0) > -0.0);
        assert!(increment_ulp(f32::MAX) == f32::INFINITY);
        assert!(increment_ulp(f32::INFINITY) == f32::INFINITY);
        assert!(increment_ulp(-f32::INFINITY) == -f32::INFINITY);
        assert!(increment_ulp(f32::NAN).is_nan());
        assert!(increment_ulp(-f32::NAN).is_nan());
    }

    #[test]
    fn dec_ulp() {
        assert!(decrement_ulp(1.0) < 1.0);
        assert!(decrement_ulp(-1.0) < -1.0);
        assert!(decrement_ulp(0.0) < 0.0);
        assert!(decrement_ulp(0.0) < -0.0);
        assert!(decrement_ulp(-0.0) < 0.0);
        assert!(decrement_ulp(-0.0) < -0.0);
        assert!(decrement_ulp(f32::MIN) == -f32::INFINITY);
        assert!(decrement_ulp(f32::INFINITY) == f32::INFINITY);
        assert!(decrement_ulp(-f32::INFINITY) == -f32::INFINITY);
        assert!(decrement_ulp(f32::NAN).is_nan());
        assert!(decrement_ulp(-f32::NAN).is_nan());
    }

    #[test]
    fn inc_dec_ulp() {
        assert_eq!(decrement_ulp(increment_ulp(0.0)), 0.0);
        assert_eq!(decrement_ulp(increment_ulp(-0.0)), 0.0);
        assert_eq!(decrement_ulp(increment_ulp(1.0)), 1.0);
        assert_eq!(decrement_ulp(increment_ulp(-1.0)), -1.0);
        assert_eq!(decrement_ulp(increment_ulp(1.2)), 1.2);
        assert_eq!(decrement_ulp(increment_ulp(-1.2)), -1.2);
    }

    #[test]
    fn dec_inc_ulp() {
        assert_eq!(increment_ulp(decrement_ulp(0.0)), 0.0);
        assert_eq!(increment_ulp(decrement_ulp(-0.0)), 0.0);
        assert_eq!(increment_ulp(decrement_ulp(1.0)), 1.0);
        assert_eq!(increment_ulp(decrement_ulp(-1.0)), -1.0);
        assert_eq!(increment_ulp(decrement_ulp(1.2)), 1.2);
        assert_eq!(increment_ulp(decrement_ulp(-1.2)), -1.2);
    }
}
